library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.process_packet.all;

-- interface do Sistema1
entity Sistema is
	port(
	clock:		in std_logic;
	sw1:		in std_logic;
-- Retira daqui
	txd:		in std_logic;
	rxd:		out std_logic;
-- ate aqui para simular	
-- Retira daqui
--	tx_av:		in std_logic;
--	rx_busy:	in std_logic;
--	rx_start:	out std_logic;
--	tx_data:	in std_logic_vector(7 downto 0);
--	rx_data:	out std_logic_vector(7 downto 0);
-- at aqui para prototipar
	display:	out std_logic_vector(13 downto 0));
end Sistema;

-- implementao do Sistema
architecture Sistema of Sistema is

-- interface do Arbitro do Barramento
component ArbitroBarramento is
	port(
	clock:		in std_logic;
	reset:		in std_logic;
	request:	in std_logic_vector(2 downto 0);
	release:	in std_logic_vector(2 downto 0);
	busy:		out std_logic;
	abort:		out std_logic;
	grant:		out std_logic_vector(2 downto 0));
end component;

-- interface do Wrapper da Blockram
component WrapperBlockram is
	port(
	clock:		in std_logic;
	reset:		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	busy:		in std_logic;
	abort:		in std_logic;
	grant:		in std_logic;
	request:	out std_logic;
	release:	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0);
	data_inMem:	in std_logic_vector(15 downto 0);
	data_outMem:	out std_logic_vector(15 downto 0);
	addressMem:	in std_logic_vector(9 downto 0);
	ce:		in std_logic;
	rw:		in std_logic;
	ack:		out std_logic);
end component;

-- interface do Wrapper da Serial
component WrapperSerial is
port(
	clock:		in std_logic;
	reset:		in std_logic;
	state:		out std_logic_vector(7 downto 0);
  	addressC:	in std_logic_vector(7 downto 0);
-- retira daqui
	txd:		in std_logic;
	rxd:		out std_logic;
-- at aqui para simular
-- retira daqui
--	rx_data:	out std_logic_vector (7 downto 0);  
--	rx_start:	out std_logic;                     
--	rx_busy:	in std_logic;                     	
--	tx_data:	in std_logic_vector (7 downto 0); 
--	tx_av:	in std_logic;                        	
-- at aqui para prototipar
	request:	out std_logic;
	grant:		in std_logic;
	release:	out std_logic;
	busy:		in std_logic;
	abort:		in std_logic;
	data_out:	out std_logic_vector(15 downto 0);
	data_in:	in std_logic_vector(15 downto 0));
end component;

-- interface do Wrapper do Processador
component WrapperProcessador is
	port(
	clock:		in std_logic;
	reset:		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	addressSerial:	in std_logic_vector(7 downto 0);
	busy:		in std_logic;
	abort:		in std_logic;
	grant:		in std_logic;
	request:	out std_logic;
	release:	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0);
	data_inMem:	in std_logic_vector(15 downto 0);
	data_outMem:	out std_logic_vector(15 downto 0);
	addressMem:	out std_logic_vector(15 downto 0);
	ce:		out std_logic;
	rw:		out std_logic;
	ack:		in std_logic);
end component;

--interface de codificao do display
component displaydecoder is
port
(
	datain: in std_logic_vector (3 downto 0);
 	dataout: out std_logic_vector (6 downto 0)
);
end component;

signal stateWS:				 	 	std_logic_vector(7 downto 0);
signal reset:						std_logic;
signal busy, abort:					std_logic;
signal grant, request, release, owner:			std_logic_vector(2 downto 0);
signal adCWS,adCWB1,adCWP1:				std_logic_vector(7 downto 0);
signal data,dataWB1,dataWB2,dataWS,dataWP1:		std_logic_vector(15 downto 0);
--BR P1
signal data_inMem_P1, data_outMem_P1:			std_logic_vector(15 downto 0);
signal addressMem_P1:					std_logic_vector(15 downto 0);
signal ce_P1, rw_P1, ack_P1:				std_logic;

begin
	reset <= not sw1;

	adCWP1 <= ADDRESSCWP1;
	adCWB1 <= ADDRESSCWB1;
	adCWS <= ADDRESSCWS;

	--display high
 	dispdechigh: displaydecoder 
 	port map(
 		datain => stateWS(7 downto 4),
 		dataout => display (13 downto 7) 
 	);
	
	--display low
	dispdeclow: displaydecoder
 	port map(
 		datain => stateWS(3 downto 0),
 		dataout => display (6 downto 0)
 	);

	
	AB : ArbitroBarramento
	port map(
		clock => clock,
		reset => reset,
		request => request,
		release => release,
		busy => busy,
		abort => abort,
		grant => grant
	);

	WS : WrapperSerial
	port map (
		clock => clock,
		reset => reset,
		state => stateWS,
		addressC => adCWS,
-- Retira daqui
		txd => txd,
		rxd => rxd,
-- at aqui para simular
-- Retira daqui
--		rx_data => rx_data,
--		rx_start => rx_start,
--		rx_busy => rx_busy,
--		tx_data => tx_data,
--		tx_av => tx_av,
-- at aqui para prototipar
		request => request(0),	   
		grant => grant(0),	   
		release => release(0),	   
		busy => busy,
		abort =>abort,
		data_out => dataWS,
		data_in => data
	);

	WP1: WrapperProcessador
	port map(
		clock => clock,
		reset => reset,
		addressC => adCWP1,
		addressSerial => adCWS,
		busy => busy,
		abort => abort,
		grant => grant(1),
		request => request(1),
		release => release(1),
		data_in => data,
		data_out => dataWP1,
		data_inMem => data_outMem_P1,
		data_outMem => data_inMem_P1,
		addressMem => addressMem_P1,
		ce => ce_P1,
		rw=> rw_P1,
		ack => ack_P1
	);
	
	WB1: WrapperBlockram
	port map(
		clock => clock,
		reset => reset,
		addressC => adCWB1,
		busy => busy,
		abort => abort,
		grant => grant(2),
		request => request(2),
		release => release(2),
		data_in => data,
		data_out => dataWB1,
		data_inMem => data_inMem_P1,
		data_outMem => data_outMem_P1,
		addressMem => addressMem_P1(9 downto 0),
		ce => ce_P1,
		rw => rw_P1,
		ack=> ack_P1
	);

	process(clock,busy,grant)
	begin
		if busy='0' then
			owner<=(others=>'0');
		elsif clock'event and clock='1' then
			if grant(0)='1' then owner(0)<='1';
			elsif grant(1)='1' then owner(1)<='1';
			elsif grant(2)='1' then owner(2)<='1';
			end if;
		end if;
	end process;
	
	data <= dataWS when owner(0)='1' else
			dataWP1 when owner(1)='1' else
			dataWB1 when owner(2)='1' else
			(others=> '1');
			
end Sistema;
