library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.process_packet.all;

-- Retira daqui
--library UNISIM;
--use UNISIM.vcomponents.all;
--use IEEE.VITAL_Timing.all;
-- at aqui para prototipar

-- interface do WrapperBlockram
entity WrapperBlockram is
	port(
	clock:			in std_logic;
	reset:			in std_logic;
	addressC:		in std_logic_vector(7 downto 0);
	busy:			in std_logic;
	abort:			in std_logic;
	grant:			in std_logic;
	data_in:		in std_logic_vector(15 downto 0);
	request:		out std_logic;
	release:		out std_logic;
	data_out:		out std_logic_vector(15 downto 0);
	data_inMem:		in std_logic_vector(15 downto 0);
	data_outMem:	out std_logic_vector(15 downto 0);
	addressMem:		in std_logic_vector(9 downto 0);
	ce:				in std_logic;
	rw:				in std_logic;
	ack:			out std_logic);
end WrapperBlockram;

-- implementao do WrapperBlockram
architecture WrapperBlockram of WrapperBlockram is

-- interface do Send
component SendReceive is
port(
	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	avSend:		in std_logic;
	nwordSend:	in std_logic_vector(2 downto 0);
	busySend:	out std_logic;
	busyWrapper:	in std_logic;
	dataSend:	in packet;
	dataReceive:	out packet;
	avReceive:	out std_logic;
	busy: 		in std_logic;
	abort: 		in std_logic;
	request: 	out std_logic;
	grant: 		in std_logic;
	release: 	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0));
end component;

-- interface do bloco de RAM
component RAMB4_S4
	port(
	WE, EN, RST, CLK:	in std_logic;
	ADDR:			in std_logic_vector(9 downto 0);
	DI:			in std_logic_vector(3 downto 0);
	DO:			out std_logic_vector(3 downto 0));
end component;

-- maquina de estados do acesso da memria pelo barramento
type stateBus is (S0,S1,S2,S3,S4,S5,S6,S7,S8);
signal ES, PES: stateBus;
signal accessBR: std_logic_vector(1 downto 0);
signal Nrw,en,we,Nclock: std_logic;
signal addressCReceive,addressCSource: std_logic_vector(7 downto 0);
signal command: std_logic_vector(15 downto 0);
signal addressBR,addr: std_logic_vector(9 downto 0);
signal din,dout,data_inBR,data_outBR: std_logic_vector(15 downto 0);

-- SendReceive
signal avSend, avReceive, busySend, busyWrapper: std_logic;
signal nwordSend: std_logic_vector(2 downto 0);
signal dataSend, dataReceive: packet;

signal 	stateWB: 		std_logic_vector(7 downto 0);
signal 	counter: 		std_logic_vector(21 downto 0);

begin

	Nclock <= clock;
	Nrw <= not rw;

	SR : SendReceive
	port map(
	clock => Nclock,
	reset => reset,
	addressC => addressC,
	avSend => avSend,
	nwordSend => nwordSend,
	busySend => busySend,
	busyWrapper => busyWrapper,
	dataSend => dataSend,
	dataReceive => dataReceive,
	avReceive => avReceive,
	busy => busy,
	abort => abort,
	request => request,
	grant => grant,
	release => release,
	data_in => data_in,
	data_out => data_out);
	
	--instanciando o bloco de RAM
	RAM0: RAMB4_S4
	port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(3 downto 0), DO=>dout(3 downto 0));
	--instanciando o bloco de RAM
	RAM1: RAMB4_S4
	port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(7 downto 4), DO=>dout(7 downto 4));
	--instanciando o bloco de RAM
	RAM2: RAMB4_S4
	port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(11 downto 8), DO=>dout(11 downto 8));
	--instanciando o bloco de RAM
	RAM3: RAMB4_S4
	port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(15 downto 12), DO=>dout(15 downto 12));

	process(reset,clock)
	begin
		if reset='1' then
			ES<=S0;
		elsif clock'event and clock='1' then
			ES<=PES;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA
	------------------------------------------------------------------------------------------------------	
	process(ES,avReceive,command,accessBR,busySend)
	begin
		case ES is
			when S0 => if avReceive='1' then PES <= S1; else PES <= S0; end if;
			when S1 => if command=x"0" and accessBR="00" then PES <= S3;
					elsif command=x"0" and accessBR/="00" then PES <= S2;
					elsif command=x"1" and accessBR="00" then PES <= S8;
					elsif command=x"1" and accessBR/="00" then PES <= S7;
					else PES<= S0; end if;
			when S2 => if accessBR= "00" then PES <= S3; else PES<= S2; end if;
			when S3 => PES<= S4;
			when S4 => PES<= S5;
			when S5 => if busySend='0' then PES <= S6; else PES<= S5; end if;
			when S6 => PES<= S0;
			when S7 => if accessBR= "00" then PES <= S8; else PES<= S7; end if;
			when S8 => PES<= S0;			
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados do acessso BARRAMENTO
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case ES is
				-- estado inicial
				when S0 =>
					addressCReceive <= (others => '0');
					addressCSource <= (others => '0');
					command <= (others => '0');
					addressBR <= (others => '0');
					data_inBR <= (others => '0');
					avSend <= '0';
					stateWB <= x"00";
					busyWrapper <= '0';
				when S1 =>
					-- "primeiro dado" [addressCReceive][addressCSource]
					addressCReceive <= dataReceive(0)(15 downto 8);
					addressCSource <=dataReceive(0)(7 downto 0);
					-- "segundo dado" [command]
					command <= dataReceive(1);
					-- "terceiro dado" [000000][addressBR]
					addressBR <= dataReceive(2)(9 downto 0);
					-- "quarto dado" [dataBR]
					data_inBR <= dataReceive(3);
					stateWB <= x"01";
					busyWrapper <= '1';
				------------------------------------
				-- Estados de LEITURA S2 a S5 
				------------------------------------
				-- esperando liberao de acesso
				when S2 =>
					avSend <= '0';
					stateWB <= x"02";
				-- lendo da blockram
				when S3 =>
					avSend <= '0';
					stateWB <= x"03";
				-- recebendo dado da leitura
				when S4 =>
					stateWB <= x"04";
				-- esperando liberao do Send
				when S5 =>
					dataSend(0) <= addressCSource & addressC;
					dataSend(1) <= command;
					dataSend(2) <= data_outBR;
					nwordSend <= "011";
					avSend <= '0';
					stateWB <= x"05";
				-- enviando
				when S6 =>
					dataSend(2) <= data_outBR;
					avSend <= '1';
					stateWB <= x"06";
					counter <= (others=> '0');
				------------------------------------
				-- Estados de ESCRITA S6 a S7
				------------------------------------
				-- esperando liberao de acesso
				when S7 =>
					avSend <= '0';
					stateWB <= x"07";
				-- escrevendo na blockram
				when S8 =>
					avSend <= '0';
					stateWB <= x"08";
			end case;
		end if;
	end process;	
	
	process(clock,reset,accessBR)
	begin
		if reset='1' then
			ack<='0';
		elsif clock'event and clock='0' then
			if accessBR="01" then
				ack<='1';
			else
				ack<='0';
			end if;
		end if;
	end process;

	process(clock,reset,accessBR)
	begin
		if reset='1' then
			data_outBR<=(others=>'0');
		elsif clock'event and clock='0' then
			if ES=S4 then
				data_outBR<= dout;
			end if;
		end if;
	end process;
	
	
accessBR <= "10" when ES=S3 or ES=S8 else --  S3 lendo da blockram e S8 escrevendo na blockram.
		"01" when ce='1' else
	    	"00"; 
			
we <= Nrw when accessBR="01" else
      '1' when ES=S8 else
      '0';
      
en <= '1' when accessBR/="00" else '0';

addr <= addressMem when accessBR="01" else
        addressBR when accessBR="10" else
        (others=>'0');

din <= data_inMem when accessBR="01" else
       data_inBR when accessBR="10" else
       (others=>'0');

data_outMem<= dout when accessBR="01" else (others=>'0');

end WrapperBlockram;
