library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
--use IEEE.VITAL_Timing.all;
use work.process_packet.all;

-- interface do WrapperProcessador
entity WrapperProcessador is
	port(
	clock:			in std_logic;
	reset:			in std_logic;
	addressC:		in std_logic_vector(7 downto 0);
	addressSerial:		in std_logic_vector(7 downto 0);
	busy:			in std_logic;
	abort:			in std_logic;
	grant:			in std_logic;
	request:		out std_logic;
	release:		out std_logic;
	data_in:		in std_logic_vector(15 downto 0);
	data_out:		out std_logic_vector(15 downto 0);
	data_inMem:		in std_logic_vector(15 downto 0);
	data_outMem:		out std_logic_vector(15 downto 0);
	addressMem:		out std_logic_vector(15 downto 0);
	ce:			out std_logic;
	rw:			out std_logic;
	ack:			in std_logic);
end WrapperProcessador;

-- implementao do WrapperProcessador
architecture WrapperProcessador of WrapperProcessador is

component processador is
port(
	ck,rst:  in std_logic;
	dataIN:  in  reg16;
	dataOUT: out reg16;
	address: out reg16;
	ce,rw:   out std_logic;
	waitR8:  in std_logic;
	haltR8:  out std_logic);
end component;

-- interface do Send
component SendReceive is
port(
	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	avSend:		in std_logic;
	nwordSend:	in std_logic_vector(2 downto 0);
	busySend:	out std_logic;
	busyWrapper:in std_logic;
	dataSend:	in packet;
	dataReceive:out packet;
	avReceive:	out std_logic;
	busy: 		in std_logic;
	abort: 		in std_logic;
	request: 	out std_logic;
	grant: 		in std_logic;
	release: 	out std_logic;
	data_in:	in std_logic_vector(15 downto 0);
	data_out:	out std_logic_vector(15 downto 0));
end component;

type state is (SP0,SP1,SP2,SP3,SP4,SP5,SP6,SP7,SP8,SP9,SP10,SP11,SP12,SP13,SP14,SP15,SP16);
signal SP, PSP: state;
signal en,write,waitR8, haltR8:	std_logic;
signal addressCReceive:		std_logic_vector(7 downto 0);
signal addressMemDist:		std_logic_vector(15 downto 0);
signal addressP,data_inP,data_outP:	std_logic_vector(15 downto 0);
signal timeout, nrepeat: std_logic_vector(3 downto 0);
signal nciclos,nciclosExecution:	std_logic_vector(31 downto 0);
signal waitfor:						std_logic_vector(7 downto 0);

-- SendReceive
signal avSend, avReceive, busySend, busyWrapper:	std_logic;
signal nwordSend:		std_logic_vector(2 downto 0);
signal dataSend, dataReceive:		packet;
signal dataOUT, dataIN:		std_logic_vector(15 downto 0);
signal Nclock:				std_logic;

begin
	Nclock <= clock;
	data_out <= dataOUT;
	dataIN <= data_in;

	rw <= write;	

	-- temporariamente
	busyWrapper <= '0';
	
	proc : processador
	port map(
		ck => clock,
		rst => reset,
		dataIN => data_inP,
		dataOUT => data_outP,
		address => addressP,
		ce => en,
		rw => write,
		waitR8 => waitR8,
		haltR8 => haltR8
	);

	SR : SendReceive
	port map(
	clock => Nclock,
	reset => reset,
	addressC => addressC,
	avSend => avSend,
	nwordSend => nwordSend,
	busySend => busySend,
	busyWrapper => busyWrapper,
	dataSend => dataSend,
	dataReceive => dataReceive,
	avReceive => avReceive,
	busy => busy,
	abort => abort,
	request => request,
	grant => grant,
	release => release,
	data_in => dataIN,
	data_out => dataOUT);


	process(reset,clock)
	begin
		if reset='1' then
			SP<=SP0;
		elsif clock'event and clock='1' then
			SP<=PSP;
		end if;
	end process;

	process(reset,clock)
	begin
		if reset='1' then
			nciclos <= (others=>'0');
		elsif clock'event and clock='1' then
			if SP/=SP0 then
				nciclos <= nciclos + '1';
			end if;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA.
	------------------------------------------------------------------------------------------------------	
	process(SP,avReceive,dataReceive,waitfor,haltR8,en,write,addressP,busySend)
	begin
		case SP is
			-- habilitacao do processador
			when SP0 => if avReceive='1' and dataReceive(1)=x"02" then PSP <= SP1; else PSP <= SP0; end if;
			-- processando
			when SP1 => PSP <= SP2;
			when SP2 => if haltR8='1' then PSP <= SP11;
						elsif en='1' and (addressP = x"FFFF") and write='0' then PSP <= SP3;
						elsif en='1' and (addressP = x"FFFF") and write='1' then PSP <= SP5;
						elsif en='1' and (addressP = x"FFFE") then PSP <= SP14;
						elsif en='1' and (addressP = x"FFFD") then PSP <= SP16;
						elsif en='1' and (addressP > x"03FF") and write='1' then PSP <= SP6;
						elsif en='1' and (addressP > x"03FF") and write='0' then PSP <= SP10;
						else PSP<= SP2;
						end if;
			-- armazena dados do printf
			when SP3 => if busySend='0' then PSP <= SP4; else PSP<= SP3; end if;
			--ativando send
			when SP4 => PSP <= SP1;
			-- armazena dados do scanf 
			when SP5 => if busySend='0' then PSP <= SP7; else PSP<= SP5; end if;
			-- armazena dados para ler da memoria distribuida
			when SP6 => if busySend='0' then PSP <= SP7; else PSP<= SP6; end if;
			-- ativando send e indo esperar dado de resposta
			when SP7 => PSP <= SP8;
			-- esperando dado
			when SP8 => if avReceive='1' and (dataReceive(1)=x"00" or dataReceive(1)=x"04") then PSP <= SP9; else PSP <= SP8; end if;
			-- recebendo dado
			when SP9 => PSP <= SP1;
			-- armazena dados para escrever em memoria distribuida
			when SP10 => if busySend='0' then PSP <= SP4; else PSP<= SP10; end if;
			-- captura o numero de ciclos de clock da execucao do processador
			when SP11 => PSP <= SP12;
			-- armazena dados do comando execution time
			when SP12 => if busySend='0' then PSP <= SP13; else PSP<= SP12; end if;
			-- ativando o send e finalizando o processamento
			when SP13 => PSP <= SP0;
			-- WAIT - armazena o addressC de que se espera o NOTIFY
			when SP14 => PSP<= SP15; 
			-- WAIT - verifica se chegou notify do addressC que se esta esperando
			when SP15 => if avReceive='1' and dataReceive(1)=x"8" and dataReceive(0)(7 downto 0)=waitfor then PSP<=SP1; else PSP<=SP15; end if; 
			-- NOTIFY - monta o pacote do notify
			when SP16 => if busySend='0' then PSP <= SP4; else PSP<= SP16; end if;

		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SP is
				when SP0 | SP1 | SP2 | SP8 | SP9 | SP15 =>
					avSend <= '0';
				-- comando Printf
				when SP3 =>
					-- numero de palavras do printf
					nwordSend <= "011";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= addressSerial & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0003";
					-- "terceiro dado" [data_outP]
					dataSend(2) <= data_outP;
				-- ativando SEND
				when SP4 | SP7 | SP13 =>
					avSend <= '1';
				-- comando Scanf
				when SP5 =>
					-- numero de palavras do scanf
					nwordSend <= "010";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= addressSerial & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0004";

				-- comando para acesso de leitura a Memoria distribuida
				when SP6 =>
					-- numero de palavras do acesso
					nwordSend <= "011";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= addressCReceive & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0000";
					-- "terceiro dado" [address]
					dataSend(2) <= addressMemDist;
				-- comando para acesso de escrita a Memoria distribuida
				when SP10 =>
					-- numero de palavras do acesso
					nwordSend <= "100";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= addressCReceive & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0001";
					-- "terceiro dado" [address]
					dataSend(2) <= addressMemDist;
					-- "quarto dado" [data]
					dataSend(3) <= data_outP;
				-- armazena o numero de ciclos de clock a execucao gastou
				when SP11 =>
					nciclosExecution<= nciclos;

				-- dados para envio do comando execution time
				when SP12 =>
					-- numero de palavras do acesso
					nwordSend <= "100";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= addressSerial & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0006";
					-- "terceiro dado" [nciclosExecution(31 downto 16)]
					dataSend(2) <= nciclosExecution(31 downto 16);
					-- "quarto dado" [nciclosExecution(31 downto 16)]
					dataSend(3) <= nciclosExecution(15 downto 0);
				-- WAIT - armazenado o addressC de que se espera o NOTIFY
				when SP14 =>
					waitfor <= data_outP(7 downto 0);
				-- NOTIFY - monta o pacote do comando notify
				when SP16 =>
					-- numero de palavras do printf
					nwordSend <= "010";
					-- "primeiro dado" [core destino][core origem]
					dataSend(0) <= data_outP(7 downto 0) & addressC;
					-- "segundo dado" [command]
					dataSend(1) <= x"0008";
			end case;
		end if;
	end process;

	process(clock,SP,en,ack,waitR8)
	begin
		if SP=SP0 then
			waitR8<='1';
		elsif clock'event and clock='0' then
			if (en ='1' and waitR8='0') or SP=SP11 or SP=SP14 then
				waitR8<= '1';
			elsif (ack='1' or SP=SP1 or SP=SP9) then
				waitR8<= '0';
			end if;
		end if;
	end process;

	addressMem <= addressP;
	data_outMem <= data_outP;
	ce <= '1' when en='1' and addressP(15 downto 10)="000000" else
		  '0';

	data_inP <= dataReceive(2) when SP=SP9 else -- retorno do scanf ou retorno de leitura a memoria distribuida
			data_inMem;


	addressMemDist <= addressP - x"0400" when (addressP>x"03FF" and addressP<x"0800" and (SP=SP6 or SP=SP10)) else
					  addressP - x"0800" when (addressP>x"07FF" and addressP<x"0C00" and (SP=SP6 or SP=SP10)) else
					  (others=>'0');
	
	addressCReceive <= x"04" when (addressP>x"03FF" and addressP<x"0800" and (SP=SP6 or SP=SP10)) else
					   x"06" when (addressP>x"07FF" and addressP<x"0C00" and (SP=SP6 or SP=SP10)) else
					   (others=>'0');
		
end WrapperProcessador;
